! VP rheology routines
! ===============================================
! The driving routine is VPdynamics; it calls  other routines.
! They include vp_matrix_rhs, vpbc and solve_vp.
! vp_matrix_rhs assembles matrix and rhs in a single cycle
! over elements.
! vpbc sets boundary conditions of no slip. Free-slip 
! boundary conditions are not implemented yet.
! solve_vp calls PETSc solver. Inside the routine there 
! are several variants for Pmode which defines how solver is working.
! It is recommended to try several choices.

!
! icestiff is used to store the velocity stiffness matrix.
! If implicit advection is selected icestiff is also used to store
! the advection stiffness matrix. Since the 
! matrices are updated on each time step there is no conflict.
 
 
! ATTENTION Metrics differentiation is introduced, but not tested
!   
! The algorithm is basically of Zhang and Hibler 1997, but the
! operator part is used as recommended in Hutchings et al (2003) 
 
! =====================================================================
SUBROUTINE VPmatrix_rhs
! Fills in the entries of the vp rheology stiffness matrix
! and simultaneously computes the rhs vector

USE ice
USE MESH
USE ELEMENTS
USE PARAM
use iceparam
USE ICE_THERM_PARMS
USE PARSUP

IMPLICIT NONE 

REAL(kind=8)                  :: m_e, a_e, drag
REAL(kind=8)                  :: dx(3),dy(3), pressure, xi, eta
REAL(kind=8)                  :: eps11,eps22,eps12, epsu, epsv, delta
REAL(kind=8)                  :: inv_mass, elex, eley, vale, entries(3)
INTEGER                       :: col, elem, elnodes(3), offset
INTEGER                       :: ipos, row, q, n, i,nini, nend
real(kind=8)                  :: ac, as, metric_f, ue, ve

ac=cos(theta_io)
as=sin(theta_io)
! The cos part is always implicit, the sin part --explicit

 do i=1,myDim_nod2D       ! Put the mass matrix part+ocean/ice friction part
                          ! + the Coriolis part. This is done in a lumped way,
			  ! so the cycle is over the owned nodes   
    row=myList_nod2D(i)
    nini=icestiff%rowptr(row)
    nend=icestiff%rowptr(row+1)-1
    inv_mass=(rhoice*m_ice(row)+rhosno*m_snow(row))
    inv_mass=max(inv_mass, 9.0_8)      ! Limit the weighted mass 
                                       ! if it is too small
    inv_mass=a_ice(row)/inv_mass
    drag=sqrt((u_ice(row)-u_w(row))**2+(v_ice(row)-v_w(row))**2)* &
         C_d_io*density_0*inv_mass
    DO q=nini,nend
     if(icestiff%colind(q)==row) then
     icestiff%values(q)=lmass_matrix(row)*(1.0/dt+drag*ac)
     else
     icestiff%values(q)=0.0
     end if
    END DO 
     rhs_u(row)=lmass_matrix(row)*(rhs_m(row)/dt+drag*ac*u_w(row) &
                +coriolis_nod2D(row)*v_ice(row)+&
                stress_atmice_x(row)*inv_mass-&
		drag*as*(v_w(row)-v_ice(row)))
     rhs_v(row)=lmass_matrix(row)*(rhs_a(row)/dt+drag*ac*v_w(row) &
                -coriolis_nod2D(row)*u_ice(row)+&
                stress_atmice_y(row)*inv_mass+&
		drag*as*(u_w(row)-u_ice(row)))
 end do
 
! Cycle over the elements to assemble the rheology part. 
! Only the owned part of the stiffness matrix is filled 
     vale=1.0_8/(ellipse**2) 
  DO n=1, myDim_elem2D     
     elem=myList_elem2D(n) 
     elnodes=elem2D_nodes(:,elem)
     
     delta=product(m_ice(elnodes))*product(a_ice(elnodes))
     if (delta==0) CYCLE       ! Skip if there is no ice
       
     dx=bafux(:,elem)
     dy=bafuy(:,elem)

	  ! Deformation rate tensor on element elem 
	  ! (differentiating metrics is omitted):
     eps11=sum(dx*u_ice(elnodes))
     eps22=sum(dy*v_ice(elnodes))
     epsu=sum(dy*u_ice(elnodes))
     epsv=sum(dx*v_ice(elnodes))
     eps12=0.5_8*(epsu+epsv)
         ! delta and mean ice properties on element:
     delta=(eps11**2+eps22**2)*(1.0_8+vale)+4.0_8*vale*eps12**2 + &
            2.0_8*eps11*eps22*(1.0_8-vale)
     delta=sqrt(delta)
     m_e=sum(m_ice(elnodes))/3.0
     a_e=sum(a_ice(elnodes))/3.0
         ! pressure and viscosities 
     pressure=0.5_8*pstar*m_e*exp(-c_pressure*(1.0_8-a_e)) 
     !xi=pressure/max(delta,delta_min)
     xi = pressure/(delta+delta_min)   ! new
     pressure=xi*delta                 ! new
     eta=xi*vale
         ! contribution from ssh
     elex=-g*sum(dx*elevation(elnodes))*voltriangle(elem)/3.0
     eley=-g*sum(dy*elevation(elnodes))*voltriangle(elem)/3.0
     metric_f=sin_elem2D(elem)/cos_elem2D(elem)/r_earth
     ue=sum(u_ice(elnodes))/3.0
     ve=sum(v_ice(elnodes))/3.0
     
     DO i=1,3  ! Cycle over rows elem contributes to
	row=elnodes(i)
	if(part(row).ne.mype) cycle
	DO q=1,nghbr_nod2D(row)%nmb
           col_pos(nghbr_nod2D(row)%addresses(q))=q
        END DO 
        inv_mass=(rhoice*m_ice(row)+rhosno*m_snow(row)) !/a_ice(row)
        inv_mass=max(inv_mass, 9.0_8)        ! Limit the weighted mass 
                                           ! if it is too small
        inv_mass=1.0/inv_mass
	
	!====== Rheology part
	                 ! u,v martices: contains phi_x(eta+xi)du/dx +
		         ! phi_y (eta+xi) du/dy; the rest is delegated to  
			 ! the rhs
			 
	offset=icestiff%rowptr(row)-1
	DO q=1,3         ! all columns 
	   col=elnodes(q)
           ipos = offset+col_pos(col)
	  
           entries(q)=(eta+xi)*(dx(i)*dx(q)+dy(i)*dy(q))* &
	              voltriangle(elem)*inv_mass  
	   icestiff%values(ipos)=icestiff%values(ipos)+entries(q)  
	END DO 	
	                 ! division by inv_mass is to be after computing
			 ! contribution from viscosities (otherwise 
			 ! information on m drops out)
			 
	                 ! u,v rhs: 
			 ! u-row: contains -eta(phi_x du/dx+
	                 ! phi_y dv/dx) + (xi/(xi+eta))*entries*u
			 ! -psi_x(xi-eta)div u +psi_x p/2
			 ! v-row: -eta (phi_x du/dy+
			 ! phi_y dv/dy)+
			 ! (xi/(xi+eta))*entries*v 
			 ! -psi_y(xi-eta)div u +psi_y p/2  
	rhs_u(row)=rhs_u(row)+(-eta*(dx(i)*eps11+dy(i)*epsv) &
	   -(xi-eta)*dx(i)*(eps11+eps22)+dx(i)*pressure)*voltriangle(elem) &
	   *inv_mass &
	   +0.8*sum(entries*u_ice(elnodes))+elex
	rhs_v(row)=rhs_v(row)+(-eta*(dx(i)*epsu+dy(i)*eps22) &
	   -(xi-eta)*dy(i)*(eps11+eps22)+dy(i)*pressure)*voltriangle(elem) &
	   *inv_mass &
	   +0.8*sum(entries*v_ice(elnodes))+eley
	                 ! 0.8=xi/(xi+eta)
	!!!!! Metric terms: They all should go to the rhs because we
	!!!!! want to keep the operator (matrix) same for u and v.

        !rhs_u(row)=rhs_u(row)+metric_f*inv_mass*(-(eta+xi)*(dx(i)*ve-epsu/3.0) &
        !   -eta*dy(i)*ue-eta*(epsv+ue*metric_f)/3.0+xi*epsu/3.0)*voltriangle(elem)	 
        
        !rhs_v(row)=rhs_v(row)+metric_f*inv_mass*(-(eta+xi)*(-dx(i)*ue-eps11/3.0 &
        !   -metric_f*ve/3.0)-xi*dx(i)*ue+(xi-eta)*(dy(i)*ve+epsv/3.0))*voltriangle(elem)	 
        rhs_u(row)=rhs_u(row)+metric_f*inv_mass*((eta+xi)*dx(i)*ve &
           -eta*dy(i)*ue-eta*(epsv+epsu+ue*metric_f)/3.0)*voltriangle(elem)	 
       
        rhs_v(row)=rhs_v(row)+metric_f*inv_mass*(-eta*ue*dx(i)+((eta+xi)*(eps11 &
           -metric_f*ve)+(xi-eta)*eps22-pressure/2.0)/3.0)*voltriangle(elem)    
     
     END DO	       
  END DO 
 
END SUBROUTINE VPmatrix_rhs
!=========================================================================
!
subroutine VPsolve

! Calls PETSc to solve parts of the VP problem. 
! Only owned rows of rhs and matrices are used
! Experimenting with Pmode1 according to the lines below may 
! speed up the performance. The comments there are obtained with 
! matrices of relatively small size and are not necessarily 
! universally valid 

use PARsup
USE MESH
USE ICE
USE PARAM
USE ICEPARAM

implicit none

#include "petscf.h"

integer        :: ident
INTEGER        :: Pmode,nrhs,restart,maxiter,lutype,fillin
INTEGER        :: COMM, myrows
REAL(kind=8)   :: droptol,soltol,rinfo(20,20)
save rinfo
INTEGER        :: Pmode1 !blocks(100)

nrhs=1
COMM=MPI_COMM_WORLD

maxiter=2000
restart=15
fillin=5
lutype=2
droptol=1.e-3
soltol=ice_VP_soltol
  ! Variants tested:
  !Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_ILU+PET_NEWPC+PET_QUIET 
  ! fastest, but less stable (requires smal soltol)  
  !Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_ILU+PET_NEWPC+PET_QUIET+PET_PCBJ
  ! fast and stable 
  !Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_PCAMG+PET_NEWPC+PET_QUIET+PET_PCBJ 
  ! reasonably fast, but slower than previous
  !Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_PCAMG+PET_NEWPC+PET_QUIET+PET_PCASM
  ! reasonably fast, similar to previous
  !Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_NEWPC+PET_QUIET+PET_PCASM
  ! faster, similar to the second in the list
  Pmode1=PET_SOLVE+PET_BICGSTAB+PET_PMVALS+PET_NEWPC+PET_QUIET+PET_PCBJ
  ! slightly faster
  !Pmode1=PET_SOLVE+PET_GMRES+PET_PMVALS+PET_NEWPC+PET_QUIET+PET_PCBJ
  ! no difference
  Pmode1=Pmode1+PET_RCM !(try on large meshes)   
   
  call MPI_Barrier(MPI_COMM_WORLD, MPIERR)
  
  Pmode=Pmode1
  if (lfirst) then
  Pmode=Pmode+PET_STRUCT
  lfirst=.false.
  end if
  call PETSC_S(Pmode,15,icestiff%dim, icestiff%nza,  myrows, &
   maxiter, & 
   restart, &  
   fillin,  &  
   droptol, &  
   soltol, &   ! Sol. Residuum reduction 
   part, icestiff%rowptr, icestiff%colind, icestiff%values, &
   rhs_u, u_ice, rinfo(:,12), COMM)
   
   Pmode=Pmode1
  call PETSC_S(Pmode,15,icestiff%dim, icestiff%nza,  myrows, &
   maxiter, & 
   restart, &  
   fillin,  &  
   droptol, &  
   soltol, &   ! Sol. Residuum reduction 
   part, icestiff%rowptr, icestiff%colind, icestiff%values, &
   rhs_v, v_ice, rinfo(:,12), COMM)

end subroutine VPsolve 
! ============================================================================
subroutine VPdynamics
USE MESH
USE ELEMENTS
USE PARAM
USE ICE
USE ICEPARAM
use PARSUP
! Driving routine of VP rheology
IMPLICIT NONE
integer      :: n, row
 DO n=1,myDim_nod2D+eDim_nod2D       
   row=myList_nod2D(n)
   rhs_m(row)=u_ice(row)
   rhs_a(row)=v_ice(row)
 END DO   
 call VPmatrix_rhs
 call VPbc
 call VPsolve
 DO n=1,myDim_nod2D+eDim_nod2D       
   row=myList_nod2D(n)
   u_ice(row)=0.5*(u_ice(row)+rhs_m(row))
   v_ice(row)=0.5*(v_ice(row)+rhs_a(row))
 END DO  
 ! Previous and predicted states are mixed as in Zhang and Hibler 
 ! 1997
 DO n=1, ice_VP_iter   !(Picard iterations)
  call VPmatrix_rhs
  call VPbc
  call VPsolve
 end DO  
 DO n=1,myDim_nod2D+eDim_nod2D       
   row=myList_nod2D(n)
   rhs_m(row)=u_ice(row)  ! It has been used to estimate drag coefficient
   rhs_a(row)=v_ice(row)  ! and of-diagonal terms (Coriolis and rotated stress)
 END DO  
 call VPsolve
 call VPmake_implicit

end subroutine VPdynamics
! ============================================================================
subroutine VPmake_implicit
USE ice
USE PARsup
USE MESH
use param
use iceparam
USE ICE_THERM_PARMS
IMPLICIT NONE 
real(kind=8)     ::  as, ac, drag, inv_mass, c1, c2, r1, r2, d
integer          ::  n, row
! updates Coriolis and drag terms to full implicit formulation
! as adviced in Zhang and Hibler 
 ac=cos(theta_io)
 as=sin(theta_io)
 DO n=1,myDim_nod2D+eDim_nod2D       
   row=myList_nod2D(n) 
   if(a_ice(row)<0.01) cycle
   inv_mass=(rhoice*m_ice(row)+rhosno*m_snow(row))/a_ice(row)
   inv_mass=max(inv_mass, 9.0)        ! Limit the weighted mass 
                                      ! if it is too small
   inv_mass=1.0/inv_mass
   drag=sqrt((rhs_m(row)-u_w(row))**2+(rhs_a(row)-v_w(row))**2)* &
         C_d_io*density_0*inv_mass
   c1=1.0_8/dt+drag*ac
   c2=coriolis_nod2D(row)+drag*as
   
   r1=c1*c1*u_ice(row)+c1*c2*(v_ice(row)-rhs_a(row))+c2*c2*rhs_m(row)
   r2=c1*c1*v_ice(row)-c1*c2*(u_ice(row)-rhs_m(row))+c2*c2*rhs_a(row)  
   d=c1*c1+c2*c2
   u_ice(row)=r1/d
   v_ice(row)=r2/d
  END DO 
end subroutine VPmake_implicit
! ============================================================================
subroutine VPbc
USE ice
USE PARsup
USE MESH
IMPLICIT NONE 

integer row, i, nini, nend, j
! Sets boundary conditions to the rhs and operator part.
DO i=1,myDim_nod2D       
   row=myList_nod2D(i)
  if((a_ice(row)<0.01).or.(bc_index_nod2D(row)<0.5)) then
  ! If there is no ice or a point is at the boundary, 
  ! enforce velocity to zero/previous value
  ! or ocean surface velocity  
   
   nini=icestiff%rowptr(row)
   nend=icestiff%rowptr(row+1)-1
   
     do j=nini, nend
       if (icestiff%colind(j)==row) then 
          icestiff%values(j)=1.0_8
       else
          icestiff%values(j)=0.0_8 
       end if
     end do
     rhs_u(row)=0._8 !u_w(row) !rhs_m(row)
     rhs_v(row)=0._8 !v_w(row) !rhs_a(row)
  end if   
END DO  


end subroutine VPbc
!=========================================================================
